/*****************************************************************************/
/*                                                                           */
/*  activePacket.c                                                           */
/*                                                                           */
/*                                                                           */
/*     Module:                                                               */
/*             activePacket.so library                                       */
/*                                                                           */
/*     Description:                                                          */
/*             Set of functions that let active aware applications build     */
/*             active packets and send them either thru globus_io or the     */
/*             standard socket API.                                          */
/*                                                                           */
/*     History:                                                              */
/*             25/06/2002 - Fabien Chanussot - Creation                      */
/*                                                                           */
/*****************************************************************************/

/*****************************************************************************/
/* Use                                                                       */
/*****************************************************************************/

#include <sys/unistd.h>
#include <netdb.h>
#include <stdio.h>
#include <errno.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <stdlib.h>
#include <sys/uio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>

#include "activePacket.h"
#include "activePacketP.h"
#include "anepP.h"


/*****************************************************************************/
/* Macros                                                                    */
/*****************************************************************************/

//Tamanoir v0.3 expects the 4bytes anep+payload size to be inserted at packet's head
#define TAMANOIR_0_3


/*****************************************************************************/
/* Types                                                                     */
/*****************************************************************************/

/*****************************************************************************/
/* Private functions                                                         */
/*****************************************************************************/

/*****************************************************************************/
/* Function: readn                                                           */
/* Role    : loop on reads until the requested amount of data is read.       */
/*****************************************************************************/
ssize_t readn(activeSocketHandle_t * hActSock, void *vptr, size_t n)
{
  size_t nleft;
  ssize_t nread;
  char *ptr;
  
  
  ptr = vptr;
  nleft = n;
  
  if ( hActSock->socket.hGlobusIo == NULL )
    { //Native sockets
      while ( nleft > 0 )
	{
	  nread = read(hActSock->socket.socket, ptr, nleft);
	  
	  if ( nread < 0 )
	    {
	      if ( errno == EINTR )
		nread = 0;
	      else
		return(-1);
	    }
	  
	  nleft -= nread;
	  ptr += nread;
	}
    }
  else
    { //globus_io
#ifdef GLOBUS_IO_SUPPORT
      globus_result_t gret;
      globus_size_t gnread;
      
      gret = globus_io_read(hActSock->socket.hGlobusIo, ptr, nleft, nleft, &gnread); //###FC - TBD: check that this works with UDP
      
      if ( gret != GLOBUS_SUCCESS )
	return(-1);
#endif
    }
  
  return(n);
}


/*****************************************************************************/
/* Protected functions                                                       */
/*****************************************************************************/

/*****************************************************************************/
/* Public functions                                                          */
/*****************************************************************************/

/*****************************************************************************/
/* Function: activeSocket                                                    */
/* Role    : turn active packets generation on for a given socket            */
/*****************************************************************************/


/****F.Bouhafs 30-12-2003********/
activeSocketHandle_t * mactiveSocket(void *socket, int type, char *service,char *destIp, int port)
{ 
  int iret;
  char optionsBuffer[IP_OPTIONS_SIZE];
  char *pBuf;
  activeSocketHandle_t *pActSock;
  socklen_t len;
  struct sockaddr_in sckAddr;
  char addrSrcStr[INET_ADDRSTRLEN];
  //char addrDestStr[INET_ADDRSTRLEN];
  //int ival;
  TLV_t TLVs[5];
  
  struct hostent *h;
  struct sockaddr_in servAddr;
  unsigned char ttl = 128;


   /* Create hActiveSocket structure */
  pActSock = (activeSocketHandle_t *)malloc(sizeof(activeSocketHandle_t));

   if ( pActSock == NULL )
    {
      activePacket_M_error(0);
      return(NULL);
    }
  
  //Init memory
  memset(pActSock, 0, sizeof(activeSocketHandle_t));
  
  //Socket descriptor
  pActSock->socketType = type;

  if (pActSock->socketType!=SOCK_DGRAM)
     {activePacket_M_error(pActSock->socketType);
      return(NULL);
     }


  //ANEP header info

  //SrcId: sender's IP (from socket)
  len = sizeof(sckAddr);
  getsockname(*((int *)socket), (struct sockaddr *)&sckAddr, &len);
  
  if ( inet_ntop(AF_INET, &sckAddr.sin_addr, addrSrcStr, INET_ADDRSTRLEN) == NULL )
    {
      activePacket_M_errno(errno);
      return(NULL);
    }
  
  TLVs[0].type = TAMANOIR_SRC_IP_TYPE;
  TLVs[0].len = strlen(addrSrcStr);
  TLVs[0].value = addrSrcStr;


  //DestId: receiver's IP from arguments

  TLVs[1].type = TAMANOIR_DEST_IP_TYPE;
  TLVs[1].len = strlen(destIp);
  TLVs[1].value = destIp;

  //LastTan: null 
  TLVs[2].type = TAMANOIR_LAST_TAN_TYPE;
  TLVs[2].len = 0;
  TLVs[2].value = NULL;

  //proxyTan: receiver's proxy TAN IP null 
  TLVs[3].type = TAMANOIR_PROXY_TAN_TYPE;
  TLVs[3].len = 0;
  TLVs[3].value = NULL;

  //serviceName: name of the service to invoke
  TLVs[4].type = TAMANOIR_SERVICE_NUMBER_TYPE;
  TLVs[4].len = strlen(service);
  TLVs[4].value = service;

  iret = anepCreate(128, 5, TLVs, &pActSock->anepHeaderSize, &pActSock->anepHeader); //###TBCheck: using default flags=128
  
  if ( iret != 0 )
    return(NULL);
    
  /* Set the Router Alert field */
  //Get the previous value
  memset(optionsBuffer, IPOPT_END, IP_OPTIONS_SIZE);
  memset(pActSock->passiveConf.ipOptions, IPOPT_END, IP_OPTIONS_SIZE);
  
  len = IP_OPTIONS_SIZE;
  iret = getsockopt(pActSock->socket.socket, IPPROTO_IP, IP_OPTIONS, (void *)optionsBuffer, &len);
  
  memcpy(pActSock->passiveConf.ipOptions, optionsBuffer, len);
  pActSock->passiveConf.ipOptionsLen = len;
  
  pBuf = optionsBuffer;
  while ( *pBuf != IPOPT_END )
    pBuf++;
  
  if ( pBuf - optionsBuffer > IP_OPTIONS_SIZE - 4 )
    {
      activePacket_M_errno(errno);
      return(NULL);
    }
  
  *pBuf = IPOPT_RA; //Option copy-class-number
  pBuf++;
  *pBuf = 4;        //Option length
  pBuf++;
  *pBuf = 0;        //Option data, cf ip_options.c
  pBuf++;
  *pBuf = 0;        //Option data, cf ip_options.c
  pBuf++;
  
  len = len + 4;
  
  while ( len & 3 )
    len++; //Align on 32bits boundary
  
  if ( len > IP_OPTIONS_SIZE )
    len = IP_OPTIONS_SIZE;
  
  iret = setsockopt(pActSock->socket.socket, IPPROTO_IP, IP_OPTIONS, (void *)optionsBuffer, len);
  //###TBD: check num version du datagramme IP (IPv4 vs IPv6).
  
  if ( iret != 0 )
    {
      activePacket_M_errno(errno);
      return(NULL);
    }
  

  h = gethostbyname (destIp);
  servAddr.sin_family = h->h_addrtype;
  servAddr.sin_port =htons(port);
  memcpy((char *) &servAddr.sin_addr.s_addr, h->h_addr_list[0],h->h_length);

  /*Check if the address is multicast*/

  if(!IN_MULTICAST(ntohl(servAddr.sin_addr.s_addr)))
     { 
        activePacket_M_errno(errno);
	return(NULL);
     }

  /*set ttl value*/
  if (setsockopt(pActSock->socket.socket,IPPROTO_IP,IP_MULTICAST_TTL,&ttl,sizeof(ttl))<0)
     {  
        activePacket_M_errno(errno);	
        return(NULL);
      }
      
  /*connect the socket to the multicast address*/
  if (connect(pActSock->socket.socket,(struct sockaddr *)&servAddr, sizeof(servAddr))<0)
    {  
        activePacket_M_errno(errno);	
        return(NULL);
    }
     
  return(pActSock);
     
}

/*****************************************************************************************************************/



activeSocketHandle_t * activeSocket(void *socket, int type, char *service, char *proxyTan, char *destIp)
{
  int iret;
  char optionsBuffer[IP_OPTIONS_SIZE];
  char *pBuf;
  activeSocketHandle_t *pActSock;
  socklen_t len;
  struct sockaddr_in sckAddr;
  char addrSrcStr[INET_ADDRSTRLEN];
  char addrDestStr[INET_ADDRSTRLEN];
  int ival;
  TLV_t TLVs[5]; 
  

#ifdef activePacket_DEBUG
  printf("### activeSocket\n");
#endif
  
  /* Create hActiveSocket structure */
  pActSock = (activeSocketHandle_t *)malloc(sizeof(activeSocketHandle_t));
  
  if ( pActSock == NULL )
    {
      activePacket_M_error(0);
      return(NULL);
    }
  
  //Init memory
  memset(pActSock, 0, sizeof(activeSocketHandle_t));
  
  //Socket descriptor
  pActSock->socketType = type;
  
  switch ( pActSock->socketType )
    {
    case NATIVE_SOCKET:
      pActSock->socket.socket = *((int *)socket);
      break;
    case GLOBUS_SOCKET:
#ifdef GLOBUS_IO_SUPPORT      
      pActSock->socket.hGlobusIo = socket;
      pActSock->socket.socket = pActSock->socket.hGlobusIo->fd;
#endif
      break;
    default:
      activePacket_M_error(pActSock->socketType);
      return(NULL);
      break;
    }
  
  //Socket type (UDP or TCP)
  len = sizeof(pActSock->streamType);
  iret = getsockopt(pActSock->socket.socket, SOL_SOCKET, SO_TYPE, (void *)&(pActSock->streamType), &len);
  
  if ( iret != 0 )
    {
      activePacket_M_errno(errno);
      return(NULL);
    }
  
  //ANEP header info
  //SrcId: sender's IP (from socket)
  len = sizeof(sckAddr);
  getsockname(*((int *)socket), (struct sockaddr *)&sckAddr, &len);
  
  if ( inet_ntop(AF_INET, &sckAddr.sin_addr, addrSrcStr, INET_ADDRSTRLEN) == NULL )
    {
      activePacket_M_errno(errno);
      return(NULL);
    }
  
  TLVs[0].type = TAMANOIR_SRC_IP_TYPE;
  TLVs[0].len = strlen(addrSrcStr);
  TLVs[0].value = addrSrcStr;
  
#ifdef activePacket_DEBUG
  printf("TLV[SrcAddr] = %s\n", addrSrcStr);
#endif
  
  //DestId: receiver's IP (from socket if UDP, from arguments if TCP)
  if ( pActSock->streamType == SOCK_STREAM )
    { //TCP
      TLVs[1].type = TAMANOIR_DEST_IP_TYPE;
      TLVs[1].len = strlen(destIp);
      TLVs[1].value = destIp;
    }
  else
    { //UDP
      len = sizeof(sckAddr);
      getpeername(*((int *)socket), (struct sockaddr *)&sckAddr, &len);
      
      if ( inet_ntop(AF_INET, &sckAddr.sin_addr, addrDestStr, INET_ADDRSTRLEN) == NULL )
	{
	  activePacket_M_errno(errno);
	  return(NULL);
	}
      
      TLVs[1].type = TAMANOIR_DEST_IP_TYPE;
      TLVs[1].len = strlen(addrDestStr);
      TLVs[1].value = addrDestStr;
      
#ifdef activePacket_DEBUG
      printf("TLV[DestAddr] = %s\n", addrDestStr);
#endif
    }
  
  
  //LastTan: null (last TAN we went thru)
  TLVs[2].type = TAMANOIR_LAST_TAN_TYPE;
  TLVs[2].len = 0;
  TLVs[2].value = NULL;
  
  //proxyTan: receiver's proxy TAN IP
  TLVs[3].type = TAMANOIR_PROXY_TAN_TYPE;
  TLVs[3].len = strlen(proxyTan);
  TLVs[3].value = proxyTan;
  
  //serviceName: name of the service to invoke
  TLVs[4].type = TAMANOIR_SERVICE_NUMBER_TYPE;
  TLVs[4].len = strlen(service);
  TLVs[4].value = service;
  
  
  iret = anepCreate(128, 5, TLVs, &pActSock->anepHeaderSize, &pActSock->anepHeader); //###TBCheck: using default flags=128
  
  if ( iret != 0 )
    {
      return(NULL);
    }
  
  /* Set the Router Alert field */
  //Get the previous value
  memset(optionsBuffer, IPOPT_END, IP_OPTIONS_SIZE);
  memset(pActSock->passiveConf.ipOptions, IPOPT_END, IP_OPTIONS_SIZE);
  
  len = IP_OPTIONS_SIZE;
  iret = getsockopt(pActSock->socket.socket, IPPROTO_IP, IP_OPTIONS, (void *)optionsBuffer, &len);
  
  memcpy(pActSock->passiveConf.ipOptions, optionsBuffer, len);
  pActSock->passiveConf.ipOptionsLen = len;
  
  pBuf = optionsBuffer;
  while ( *pBuf != IPOPT_END )
    pBuf++;
  
  if ( pBuf - optionsBuffer > IP_OPTIONS_SIZE - 4 )
    {
      activePacket_M_errno(errno);
      return(NULL);
    }
  
  *pBuf = IPOPT_RA; //Option copy-class-number
  pBuf++;
  *pBuf = 4;        //Option length
  pBuf++;
  *pBuf = 0;        //Option data, cf ip_options.c
  pBuf++;
  *pBuf = 0;        //Option data, cf ip_options.c
  pBuf++;
  
  len = len + 4;
  
  while ( len & 3 )
    len++; //Align on 32bits boundary
  
  if ( len > IP_OPTIONS_SIZE )
    len = IP_OPTIONS_SIZE;
  
  iret = setsockopt(pActSock->socket.socket, IPPROTO_IP, IP_OPTIONS, (void *)optionsBuffer, len);
  //###TBD: check num version du datagramme IP (IPv4 vs IPv6).
  
  if ( iret != 0 )
    {
      activePacket_M_errno(errno);
      return(NULL);
    }
  
  /* Set the "do not fragment" IP field */
  len = sizeof(pActSock->passiveConf.df);
  iret = getsockopt(pActSock->socket.socket, IPPROTO_IP, IP_MTU_DISCOVER, (void *)&(pActSock->passiveConf.df), &len);
  
  if ( iret != 0 )
    {
      activePacket_M_errno(errno);	
      return(NULL);
    }
  
  optionsBuffer[0] = IP_PMTUDISC_DO;
  iret = setsockopt(pActSock->socket.socket, IPPROTO_IP, IP_MTU_DISCOVER, (void *)optionsBuffer, 1);
  
  if ( iret != 0 )
    {
      activePacket_M_errno(errno);
      return(NULL);
    }
  
  
  /* Set the "TCP_NODELAY" TCP field */
  if ( pActSock->streamType == SOCK_STREAM )
    {
      int icork;
      int inodelay;
      
      len = sizeof(icork);
      iret = getsockopt(pActSock->socket.socket, IPPROTO_TCP, TCP_CORK, (void *)&icork, &len);
      
      if ( iret != 0 )
	{
	  activePacket_M_errno(errno);	
	  return(NULL);
	}
      
      len = sizeof(inodelay);
      iret = getsockopt(pActSock->socket.socket, IPPROTO_TCP, TCP_NODELAY, (void *)&inodelay, &len);
      
      if ( iret != 0 )
	{
	  activePacket_M_errno(errno);	
	  return(NULL);
	}
      
      if ( icork )
	pActSock->passiveConf.noDelay = 2;
      else if ( inodelay )
	pActSock->passiveConf.noDelay = 1;
      else 
	pActSock->passiveConf.noDelay = 0;

      if ( icork )
	{
	  ival = 0;
	  iret = setsockopt(pActSock->socket.socket, IPPROTO_TCP, TCP_CORK, (void *)&ival, sizeof(ival));
	  
	  if ( iret != 0 )
	    {
	      activePacket_M_errno(errno);
	      return(NULL);
	    }
	}
      
      ival = 1;
      iret = setsockopt(pActSock->socket.socket, IPPROTO_TCP, TCP_NODELAY, (void *)&ival, sizeof(ival));
      
      if ( iret != 0 )
	{
	  activePacket_M_errno(errno);
	  return(NULL);
	}
    }
  
  
  /* Done. */
  return(pActSock);
}


/*****************************************************************************/
/* Function: activeClose                                                     */
/* Role    : turn active packets generation off for a given socket           */
/*****************************************************************************/
int activeClose(activeSocketHandle_t *hActSock)
{
  int iret;
  char optionsBuffer[IP_OPTIONS_SIZE];
  int ival;
  
  
#ifdef activePacket_DEBUG
  printf("### activeClose\n");
#endif
  
  /* If hActSock is valid */
  if ( hActSock != NULL )
    {
      /* Clear the Router Alert field */
      memset(optionsBuffer, IPOPT_END, IP_OPTIONS_SIZE);
      
      iret = setsockopt(hActSock->socket.socket, IPPROTO_IP, IP_OPTIONS, (void *)hActSock->passiveConf.ipOptions,
			hActSock->passiveConf.ipOptionsLen);
      //###TBD: in the case of TCP, there might still be some bytes in the buffer waiting to be sent;
      //        so we should flush before unsetting the router alert...
      
      if ( iret != 0 )
	{
	  activePacket_M_errno(errno);
	  return(-1);
	}
      
      /* Restore the "do not fragment" IP field */
      iret = setsockopt(hActSock->socket.socket, IPPROTO_IP, IP_MTU_DISCOVER, (void *)&(hActSock->passiveConf.df),
			sizeof(hActSock->passiveConf.df));
      
      /* Restore the "TCP_NODELAY" TCP field */
      if ( hActSock->streamType == SOCK_STREAM )
	{
	  switch ( hActSock->passiveConf.noDelay )
	    {
	    case 0:
	      //TCP_NODELAY back to 0
	      ival = 0;
	      iret = setsockopt(hActSock->socket.socket, IPPROTO_TCP, TCP_NODELAY, (void *)&(ival),
				sizeof(ival));
	      
	      if ( iret != 0 )
		{
		  activePacket_M_errno(errno);
		  return(-1);
		}
	      break;
	      
	    case 1:
	      //Do nothing, that's the configuration we are in
	      break;
	      
	    case 2:
	      //TCP_NODELAY back to 0, TCP_CORK back to 1
	      ival = 0;
	      iret = setsockopt(hActSock->socket.socket, IPPROTO_TCP, TCP_NODELAY, (void *)&(ival),
				sizeof(ival));
	      
	      if ( iret != 0 )
		{
		  activePacket_M_errno(errno);
		  return(-1);
		}
	      
	      ival = 1;
	      iret = setsockopt(hActSock->socket.socket, IPPROTO_TCP, TCP_CORK, (void *)&(ival),
				sizeof(ival));
	      
	      if ( iret != 0 )
		{
		  activePacket_M_errno(errno);
		  return(-1);
		}
	      break;
	      
	    default:
	      activePacket_M_errno(errno);
	      return(-1);
	      break;
	    }
	}
      
      /* Free resources */
      anepDelete(hActSock->anepHeader);
      
      free(hActSock);
    }
  else
    {
      activePacket_M_error(0);
      return(-1);
    }
  
  
  /* Done. */
  return(0);
}


/*****************************************************************************/
/* Function: activeWrite                                                     */
/* Role    : Send an active packet thru the active socket                    */
/*****************************************************************************/
int activeWrite(activeSocketHandle_t *hActSock, char *payload, size_t size)
     //###TBD: return warning when fragmentation has to occur so that apps can 
     //###TBD  reduce the size of their packets.
{
  int iret;
#ifdef TAMANOIR_0_3
  struct iovec iov[3];
  uint32_t totalLength;
#else
  struct iovec iov[2];
#endif
  
  
#ifdef activePacket_DEBUG
  printf("### activeWrite\n");
#endif
  
  /* Update ANEP header so that it reflects the total packet size */
  anepSetPacketLength(hActSock->anepHeader, size);
  
  /* Construct the iov record... */
#ifdef TAMANOIR_0_3
  totalLength = htonl(hActSock->anepHeaderSize + size);
  
  iov[0].iov_base = (char *) &totalLength; //First part of the ANEP capsule
  iov[0].iov_len = sizeof(uint32_t);
  
  iov[1].iov_base = hActSock->anepHeader; //First part of the ANEP capsule
  iov[1].iov_len = hActSock->anepHeaderSize;
  
  iov[2].iov_base = payload;              //The payload
  iov[2].iov_len = size;
#else
  iov[0].iov_base = hActSock->anepHeader; //First part of the ANEP capsule
  iov[0].iov_len = hActSock->anepHeaderSize;
  
  iov[1].iov_base = payload;              //The payload
  iov[1].iov_len = size;
#endif  
  
  /* ...and send it */
  if ( hActSock->socketType == NATIVE_SOCKET )
    { //Using native sockets
      errno = 0;
#ifdef TAMANOIR_0_3
      iret = writev(hActSock->socket.socket, iov, 3);
      
      if ( iret != (iov[0].iov_len + iov[1].iov_len + iov[2].iov_len) )
	{
	  activePacket_M_error(0);
	  return(-1);
	}
#else
      iret = writev(hActSock->socket.socket, iov, 2);
      
      if ( iret != (iov[0].iov_len + iov[1].iov_len) )
	{
	  activePacket_M_error(0);
	  return(-1);
	}
#endif
    }
  else
    { //Using globus_io
#ifdef GLOBUS_IO_SUPPORT
      globus_result_t gret;
      globus_size_t bytes_written;
      
      
      if ( hActSock->streamType == SOCK_STREAM )
	{ //TCP
#ifdef TAMANOIR_0_3
	  gret = globus_io_writev(hActSock->socket.hGlobusIo, iov, 3, &bytes_written);
	  
	  if ( ( gret != GLOBUS_SUCCESS ) || ( bytes_written != iov[0].iov_len + iov[1].iov_len + iov[2].iov_len ) )
	    {
	      activePacket_M_error(0);
	      return(-1);
	    }
#else
	  gret = globus_io_writev(hActSock->socket.hGlobusIo, iov, 2, &bytes_written);
	  
	  if ( ( gret != GLOBUS_SUCCESS ) || ( bytes_written != iov[0].iov_len + iov[1].iov_len ) )
	    {
	      activePacket_M_error(0);
	      return(-1);
	    }
#endif
	}
      else
	{ //UDP
	  globus_size_t gbuffLen;
	  char *gbuffer;
	  
	  //Merge the iov into a single buffer...
#ifdef TAMANOIR_0_3
	  gbuffLen = iov[0].iov_len + iov[1].iov_len + iov[2].iov_len;
#else
	  gbuffLen = iov[0].iov_len + iov[1].iov_len;
#endif
	  
	  gbuffer = (char *)malloc(gbuffLen);
	  
	  if ( gbuffer == NULL )
	    {
	      activePacket_M_error(0);
	      return(-1);
	    }
	  
	  memcpy(gbuffer, iov[0].iov_base, iov[0].iov_len);
	  memcpy(gbuffer + iov[0].iov_len, iov[1].iov_base, iov[1].iov_len);
#ifdef TAMANOIR_0_3
	  memcpy(gbuffer + iov[0].iov_len + iov[1].iov_len, iov[2].iov_base, iov[2].iov_len);
#endif
	  
	  //... and send it
	  gret = globus_io_send(hActSock->socket.hGlobusIo, gbuffer, gbuffLen, 0, &bytes_written);

	  if ( ( gret != GLOBUS_SUCCESS ) || ( bytes_written != gbuffLen ) )
	    {
	      activePacket_M_error(0);
	      return(-1);
	    }
	  
	  free(gbuffer);
	}
#endif
    }
  
  //###FC - TBD: do that based on the socket type;
  //###FC - TBD: check below
#if 0  
  if ( iret == -1 )
    {
      if ( errno == 90 )
	{
	  /* Avoid IP fragmentation by sending two messages */
	  if ( size >= 2 )
	    {
	      iret =  activeWrite(hActSock, payload, size/2);
	      iret2 =  activeWrite(hActSock, payload + size/2, size - size/2);
	      
	      if ( ( iret != 0 ) || ( iret2 != 0 ) )
		return(-1);
	    }
	  else
	    {
	      activePacket_M_errno(errno);
	      return(-1);
	    }
	}
      else
	{
	  activePacket_M_errno(errno);
	  return(-1);
	}
    }
#endif
  
  
  /* Done */
  return(0);
}


/*****************************************************************************/
/* Function: activeRead                                                      */
/* Role    : Receive an active packet from the active socket                 */
/*****************************************************************************/
int activeRead(activeSocketHandle_t * hActSock, char *payload, size_t *size)
{
  char buffer[ANEP_LENGTH_WO_TLV];
  char *buffTmp;
  uint16_t payLoadOffset;
  uint16_t payLoadSize;
  uint16_t readSize;
  int iret;
#ifdef TAMANOIR_0_3
  uint32_t totalLength;
#endif
  
#ifdef activePacket_DEBUG
  printf("### activeRead\n");
#endif
  
  
  /* Check the socket type */
  if ( hActSock->streamType == SOCK_STREAM )
    { //TCP socket: we can do several reads on the same packet
      /* If it's a new read of a new packet */
      if ( hActSock->leftToRead == 0 )
	{
#ifdef TAMANOIR_0_3
	  //Throw length away
	  iret = readn(hActSock, (char *) &totalLength, 4);
	  
	  if ( iret == -1 )
	    {
	      activePacket_M_errno(errno);
	      return(-1);
	    }
#endif
	  
	  /* Extract ANEP header from the socket */
	  iret = readn(hActSock, buffer, ANEP_LENGTH_WO_TLV);
	  
	  if ( iret == -1 )
	    {
	      activePacket_M_errno(errno);
	      return(-1);
	    }
	  
	  //###FC - Useless if ( iret < ANEP_LENGTH_WO_TLV )
	  //###FC - Useless   { //This is not an ANEP packet
	  //###FC - Useless     memcpy(payload, buffer, iret);
	  //###FC - Useless    *size = iret;
	      
	  //###FC - Useless     activePacket_M_error(iret);	      
	  //###FC - Useless     return(-2); //###TBD: define error "not an ANEP paquet"
	  //###FC - Useless   }
	  
	  //Find out where the payload starts, and what is its size
	  iret = anepParse(buffer, &payLoadOffset, &payLoadSize);
	  
#ifdef activePacket_DEBUG
	  printf("### ANEP: payload offset %d, payload size %d\n", payLoadOffset, payLoadSize);
#endif
	  
	  if ( iret != 0 )
	    { //This is not an ANEP packet
	      memcpy(payload, buffer, iret);
	      *size = iret;
	      
	      activePacket_M_error(iret);	
	      return(-2); //###TBD: define error "not an ANEP paquet"
	    }
      
	  //Throw the TLVs part of the ANEP header away
	  if ( payLoadOffset != 0 )
	    {
	      buffTmp = (char *)malloc(payLoadOffset);
	      
	      if ( buffTmp == NULL )
		{
		  activePacket_M_error(0);	
		  return(-1);
		}
	      
	      iret = readn(hActSock, buffTmp, payLoadOffset);
	      
	      if ( iret == -1 )
		activePacket_M_errno(errno);

	      //###FC - Useless if ( iret != payLoadOffset )
	      //###FC - Useless {
	      //###FC - Useless if ( iret == -1 )
	      //###FC - Useless activePacket_M_errno(errno);
	      //###FC - Useless   else
	      //###FC - Useless     activePacket_M_error(iret);
	      //###FC - Useless   return(-1);
	      //###FC - Useless }
	      
	      free(buffTmp);
	    }
	  
	  /* Read the payload */
	  readSize = ( payLoadSize > *size ) ? *size : payLoadSize;
	  iret = readn(hActSock, payload, readSize);
	  
	  if ( iret == -1 )
	    activePacket_M_errno(errno);

	  //###FC - Uselessif ( iret != readSize )
	  //###FC - Useless  {
	  //###FC - Useless    if ( iret == -1 )
	  //###FC - UselessactivePacket_M_errno(errno);
	  //###FC - Useless   else
	  //###FC - UselessactivePacket_M_error(iret);
	  //###FC - Useless   return(-1);
	  //###FC - Useless  }
	  
	  /* Let the client know if there is still some payload to read */
	  hActSock->leftToRead = payLoadSize - readSize;
	  *size = payLoadSize;
	}
      else /* It's a read to keep on reading the payload */
	{
	  readSize = ( hActSock->leftToRead > *size ) ? *size : hActSock->leftToRead;
	  iret = readn(hActSock, payload, readSize);
	  
	  if ( iret == -1 )
	    activePacket_M_errno(errno);

	  //###FC - Uselessif ( iret != readSize )
	  //###FC - Useless  {
	  //###FC - Useless    if ( iret == -1 )
	  //###FC - UselessactivePacket_M_errno(errno);
	  //###FC - Useless    else
	  //###FC - UselessactivePacket_M_error(iret);
	  //###FC - Useless    return(-1);
	  //###FC - Useless  }
	  
	  /* Let the client know if there is still some payload to read */
	  *size = hActSock->leftToRead;
	  hActSock->leftToRead -= readSize;
	}
    }
  
  else
    //UDP socket: we've got to read the packet at once
    { 
      // Allocate ethernet MTU bytes of memory
      buffTmp = (char *) malloc(DEFAULT_DATAGRAM_SIZE); //This should really be 1500 - header size (1460?)...
      
      if ( buffTmp == NULL )
	{
	  activePacket_M_error(0);
	  return(-1);
	}
      
      // Read the datagram
      iret = read(hActSock->socket.socket, buffTmp, DEFAULT_DATAGRAM_SIZE);
      
      if ( iret == -1 )
	{
	  activePacket_M_errno(errno);
	  return(-1);
	}
      
      //Find out where the payload starts, and what is its size
#ifdef TAMANOIR_0_3 
      iret = anepParse(buffTmp + 4, &payLoadOffset, &payLoadSize);
#else
      iret = anepParse(buffTmp, &payLoadOffset, &payLoadSize);
#endif
      
#ifdef activePacket_DEBUG
      printf("### ANEP: payload offset %d, payload size %d\n", payLoadOffset, payLoadSize);
#endif
      
      if ( iret != 0 )
	{ //This is not an ANEP packet
	  memcpy(payload, buffTmp, iret);
	  *size = iret;
	  
	  free(buffTmp);
	  
	  activePacket_M_error(0);	
	  return(-2); //###TBD: define error "not an ANEP paquet"
	}
      
      if ( payLoadSize > *size )
	{
	  free(buffTmp); 
           
          activePacket_M_error(0);
	  return(-3);
	}

#ifdef TAMANOIR_0_3 
      memcpy(payload, buffTmp + payLoadOffset + ANEP_LENGTH_WO_TLV + 4, payLoadSize);
#else
      memcpy(payload, buffTmp + payLoadOffset + ANEP_LENGTH_WO_TLV, payLoadSize);
#endif
      *size = payLoadSize;
    }
  
  
  /* Done. */
  return(0);
}
