package ibp;
//
// Session.java
//
// Written by Yong Zheng.  May 2001
//

import java.util.*;
import java.net.*;
import java.io.*;


class IBPUtil {
	static IBPException getException( int errorCode) {
		return( new IBPException());
	}
}


interface IBPCommUnit {
	void execute( IBPSession session ) throws IOException,IBPException;
}

/** 
   * IBPSession represents a session between IBP client and server
   *
   *  @author	Yong Zheng
   */

class IBPSession {
	//constructor
	IBPSession( InetAddress addr, int port , int timeout ) throws
							IBPException{
		address = addr;
		this.port = port;
		setTimeout(timeout);
		begTime = new Date().getTime();
	}
	
	IBPSession( IBPDepot dpt, int timeout) throws IBPException {
		address = dpt.getAddress();
		this.port = dpt.getPort();
		setTimeout(timeout);
		begTime = new Date().getTime();
	}
	
	//fields
	private InetAddress address;
	private int	port;
	private int	timeout = 0;
	private boolean timedSession = false;
	private Socket	dataSock;
	private Socket	controlSock;
	private int	dataLen;
	private String 	result;
	private long	begTime;
	private LinkedList	list = new LinkedList();
	private DataInputStream in;
	private DataOutputStream out;
	
	//methods
	void putCommUnit( IBPCommUnit commUnit ){
		list.addFirst(commUnit);
	}
	
	void setTimeout( int timeout) throws IBPException {
		if ( timeout < 0 )
			throw new IBPIllegalArgumentException();
		this.timeout = timeout * 1000;
		timedSession = true;
		if ( timeout == 0 )
			timedSession = false;	
			
	}
	
	boolean isTimedSession( ){
		return ( timedSession);
	}
	
	String getResult( ){
		return result;
	}
	
	
	void sendRequest ( String cmd ) throws IOException{
		byte[] data = cmd.getBytes();
		out.write(data,0,data.length);
		out.flush();
	}
	
	void sendData ( byte[] buf, int offset, int size) throws IOException,IBPException{
		out.write(buf,offset,size);
		out.flush();
	}
	
	String recResponse() throws IOException,IBPException{
	
		try {   
			int delay = getTimeout();
			if ( isTimedSession()){
				if ( delay <= 0 )
					throw new InterruptedIOException();
				getControlSocket().setSoTimeout(delay);		
			}else {
				getControlSocket().setSoTimeout(0);
			}
			
			StringBuffer cmd = new StringBuffer(20);
			char lineSep = System.getProperty("line.separator").charAt(0);
			char b;
			while ( (b = (char)in.readByte()) != lineSep){
				cmd.append(b);	
			}	
			return(cmd.toString());
			
			//return in.readLine();
		}catch ( SocketException e){
			throw new IOException();
		}catch ( InterruptedIOException e){
			throw new IBPTimeoutException();
		}
		
	}
	
	int recData( byte[] buf, int offset,int length) throws IOException,IBPException{
		try {   
			int delay = getTimeout();
			if ( isTimedSession()){
				if ( delay <= 0 )
					throw new InterruptedIOException();
				getDataSocket().setSoTimeout(delay);		
			}else {
				getDataSocket().setSoTimeout(0);
			}	

			if ( length < 0)
				length = getDataLen();
			int len = in.read(buf,offset,length);
			if (len == -1 ){ 
				throw new IOException();
			}
			return(len);
		}catch ( SocketException e){
			throw new IOException();
		}catch ( InterruptedIOException e){
			throw new IBPTimeoutException();
		}
	}
	
	void perform( ) throws IOException,IBPException {
		IBPCommUnit 	unit;
		
		controlSock = createSocket(address,port);
		dataSock = controlSock;
		in = new DataInputStream(controlSock.getInputStream());
		out = new DataOutputStream( controlSock.getOutputStream());	
		try{
			if ( isTimedSession() && timeout <= 0 )
				throw new InterruptedIOException();
			while( !list.isEmpty()){
				unit = getNextUnit();
				unit.execute(this);
				if ( isTimedSession() ){
					setTimeout( timeout - (int)(begTime - new Date().getTime()));
					if ( timeout < 0)
						throw new InterruptedIOException();
				}			
			}
		}catch ( InterruptedIOException e){
			throw new IBPTimeoutException();
		}finally {
			try {
				controlSock.close();
			}catch ( IOException e ) {}
			try {
				dataSock.close();
			}catch ( IOException e) {}
		}
	}

	Socket getDataSocket(){
		return(controlSock);
	}
	
	Socket getControlSocket(){
		return(controlSock);
	}
	
	int getTimeout() {
		return(timeout);
	}
	
	int getDataLen () {
		return ( dataLen);
	}
	
	void setDataLen( int len){
		dataLen = len;
	}
	
	void setResult( String str){
		result = str;
	}
	
	private IBPCommUnit getNextUnit(){
		return (IBPCommUnit)list.removeLast();
	}
	
	private Socket createSocket(InetAddress addr, int port) throws
						IOException,IBPException{
		try {
			if ( timeout == 0 ){
				return( new Socket(addr,port));
			}else{
				Socket sock = 
				      TimedSocket.getSocket(addr,port,timeout);	
				setTimeout( timeout - 
					(int)(begTime - new Date().getTime()));
				return(sock);	
			}
		}catch ( InterruptedIOException e){
			throw new IBPTimeoutException();
		}		
	}
	
	
}

/** 
   * Request represents a basic unit of IBP communication session 
   * which is sent to ibp server.
   * The format of request is |request| ending with "\n"
   *
   *  @author	Yong Zheng
   */

class Request implements IBPCommUnit {
	// constructor
	Request( String cmd ){
		command = cmd;
	}
	
	//class fields
	private String command;
		
	//methods
	public void execute( IBPSession session) throws IOException,IBPException{
		session.sendRequest(command);
	} 
}

/** 
   * outData represents a basic unit of IBP communication session 
   * which sends use data to ibpserver.
   *
   *  @author	Yong Zheng
   */
class outData implements IBPCommUnit{
	// constructor
	outData( byte[] data, int len){
		this.data = data;
		length = len;
	}
	outData( byte[] data, int len, int offset){
		this(data,len);
		this.offset = offset;
	}
		
	// class fields
	private byte[] 	data;
	private int	offset = 0;
	private int 	length;
	
	// methods
	public void execute ( IBPSession session) throws IOException,IBPException {
		session.sendData(data,offset,length);
	}
}

/** 
   * inData represents a basic unit of IBP communication session 
   * which sends use data to ibpserver.
   *
   *  @author	Yong Zheng
   */
class inData implements IBPCommUnit {
	//constructor
	inData( byte[] data){ this.data = data;}
	inData( byte[] data, int off){
		this.data = data;
		offset  = off;
	}
	inData( byte[] data, int off, int len){
		this(data,off);
		length = len;
	}
	
	//class fields
	private int length = -1;
	private byte[] data;
	private int offset = 0;
	
	//methods
	public void execute ( IBPSession session) throws IOException,IBPException {
		int len = session.recData(data,offset,length);
		session.setDataLen(len);
		
		//System.out.println("Receive data:"+new String(data));
	}	
}




/** 
   * Response represents a basic unit of IBP communication session 
   * which is expected from ibp server.
   * The expected format is |error_code| ending with "\n"
   *
   *  @author	Yong Zheng
   */
class Response implements IBPCommUnit {
	// constructor
	Response() { }
	Response( int value ){ expectedValue = value; }
	
	// class fields
	private int	expectedValue = IBPProtocol.IBP_OK;
		
	// methods
	public void execute( IBPSession session) throws IOException, IBPException {
		int	returnValue;
		
		String  str = session.recResponse();
				
		try {
			returnValue = Integer.parseInt(str.trim());
		}catch ( NumberFormatException e ){
			throw new IBPResponseFormatException();
		}
		if ( returnValue != expectedValue ){
			throw IBPUtil.getException(returnValue); 
		}
	}
}


/** 
   * Response1 represents a basic unit of IBP communication session 
   * which is expected from ibp server.
   * The expected format is |response| ending with "\n"
   *
   *  @author	Yong Zheng
   */

class Response1 implements IBPCommUnit {
	// constructor
	
	//class fields
	
	//methods
	public void execute( IBPSession session ) throws IOException,IBPException{
		String 	str = session.recResponse();
		session.setResult(str);  
	}
}



/** 
   * Response2 represents a basic unit of IBP communication session 
   * which is expected from ibp server.
   * The expected format is |errorCode dataLen| ending with "\n"
   *
   *  @author	Yong Zheng
   */
class Response2 implements IBPCommUnit {
	// constructor
	Response2() {}
	Response2( int value ){ expectedValue = value;}
	
	// class fields
	private int expectedValue = IBPProtocol.IBP_OK;
	
	//methods
	public void execute ( IBPSession session) throws IOException, IBPException{
		String str = session.recResponse();

		//System.out.println("rece str:"+str);		
		StringTokenizer parser = 
			new StringTokenizer(str," \n");
		try {
			String str1 = parser.nextToken();
			int returnValue = Integer.parseInt(str1); 
			if ( returnValue != expectedValue )
				throw IBPUtil.getException(returnValue);
			str1 = parser.nextToken();
			int len = Integer.parseInt(str1);
			session.setDataLen(len); 
		}catch( NumberFormatException e){
			throw new IBPResponseFormatException();
		}catch( NoSuchElementException e){
			throw new IBPResponseFormatException();
		}
		/*
		try{
			str = rd.readLine();
			System.out.println("received Data:"+str);
		}catch (Exception e) {}
		*/	
		 	   
	}
}
