package ibp;

import java.util.*;
import java.net.*;
import java.io.*;

//
// ReadCap.java
//
// Written by Yong Zheng

/**
  * This class presents a IBP Capability which has read right  
  * 
  * 
  * @author   Yong Zheng
  */

public class ReadCap extends Capability {
	// constructor
	
	/**
	  * creates a read capability
	  *
	  * 	@param	cap	cap in string
	  */
	ReadCap( String cap ) throws IBPException, UnknownHostException {
		super(cap);
		if ( getType().compareTo("READ") != 0 )
			throw new IBPAccessDeniedException();
	};	  
	  
	// methods
	/**
	  * read reads up to size bytes,starting at offset from
	  * storage area.
	  *
	  *	@param	buf	data buffer
	  *	@param	size	size
	  *	@param	offset	start position
	  *	@return the number of bytes received
	  *	@exception throws IOException ,IBPException
	  */
	public long read( byte[] buf, long size, long offset) throws IOException,IBPException {
		// check parameter
		if ( size <= 0 )
			throw new IBPIllegalArgumentException();
		// create a transaction
		IBPSession session = new IBPSession(getDepot(),getClientTimeout());
		// make a request
		String cmd = ""+IBPProtocol.IBP_v031+" "+
			     IBPProtocol.IBP_LOAD+" "+
			     getName() + " "+
			     getKey() + " "+
			     offset + " "+
			     size + " " +
			     getServerTimeout() + " \n";
		session.putCommUnit(new Request(cmd));
		session.putCommUnit(new Response2());
		session.putCommUnit(new inData(buf));
		//session.putCommUnit(new Response1());
		session.perform();
		//System.out.println("result:"+session.getResult());
		//System.exit(-1);
		return( session.getDataLen());	
					
	};

	/**
	  * copy copies up to size bytes,starting at offset from
	  * storage area accessed throught IBP read capability source
	  * to target write capability.
	  *
	  *	@param	target	target capability
	  *	@param	size	size
	  *	@param	offset	start position
	  *	@param  tgtCliTm server side client timeout
	  *	@param	tgtSvrTm remote server timeout
	  *	@return the number of bytes copied
	  *	@exception throws IOException ,IBPException
	  */
	public long copyTo( WriteCap target, long size, long offset, int tgtCliTm, 
					int tgtSvrTm ) throws IOException,IBPException {
		// check parameter
		if ( size <= 0 )
			throw new IBPIllegalArgumentException();
		// create a transaction
		IBPSession session = new IBPSession(getDepot(),getClientTimeout());
		// make a request
		//System.out.print("wc="+target);
		String cmd = ""+IBPProtocol.IBP_v031+" "+
			     IBPProtocol.IBP_SEND+" "+
			     getName() + " "+
			     target + " " + 
			     getKey() + " "+
			     offset + " "+
			     size + " " +
			     getServerTimeout() + " " +
			     tgtCliTm + " "+
			     tgtSvrTm + " \n";
		session.putCommUnit(new Request(cmd));
		session.putCommUnit(new Response2());
		session.perform();
		return( session.getDataLen());	
	};						    

	/**
	  * copy copies up to size bytes,starting at offset from
	  * storage area accessed throught IBP read capability source
	  * to target write capability.
	  *
	  *	@param	target	target capability
	  *	@param	size	size
	  *	@param	offset	start position
	  *	@param  tgtCliTm server side client timeout
	  *	@param	tgtSvrTm remote server timeout
	  *	@return the number of bytes copied
	  *	@exception throws IOException ,IBPException
	  */
	public long copyTo( String target, long size, long offset, int tgtCliTm, 
				int tgtSvrTm ) throws IOException,IBPException {
		return copyTo( new WriteCap(target),size,offset,tgtCliTm,tgtSvrTm);
	};

};
